// Implementation of the routines of the object 'observe'. This object
// is fairly similar to the one defined in STOKES and manages a web of
// detectors. The program ANALYZE will store the sum of all data files
// in these detectors and finally save their contents into a new set
// of STOKES files and in convenient output files for the polarization
// data.
//
// ANALYZE, version 1.2, December 2012


#include <iostream>
#include <fstream>
#include <math.h>
#include "observe-analyze-v1.2.h"

using namespace std;



// definition of pi
const float pi=3.14159265358979323846264338327950;


// initialize a new dynamical array of detectors by setting all
// channels to zero
Monitor::Monitor()
{
  Col = new Column[ScreenNumThetaMax];
}


// add 'value' to the channel l of the detector given by the theta and
// phi position i and j; n denotes the number of the Stokes parameter
void Monitor::AddTo(int i,int j,int n,int l,int x, int y,double value)
{
  Col[i].Row[j].AddTo(n,l,x,y,value);
}


// save the data of all detectors to new combined STOKES files and to
// convenient output files
void Monitor::SaveResults(char outfile[300],char resfile[300],
                          int n,double Data[])
{
  int i,j,l,x,y,PhiNum;
  double k,dk,N,p,theta;
  double I[ResMax][ResXmax][ResYmax],Q[ResMax][ResXmax][ResYmax],U[ResMax][ResXmax][ResYmax],V[ResMax][ResXmax][ResYmax];
  ofstream out_stream,res_stream;

  // open the files
  out_stream.open(outfile,ios::app);
  res_stream.open(resfile,ios::app);

  if (Data[38])
  {
    PhiNum = 1;
  }
  else
  {
    PhiNum = Data[34];
  }
  
  // loop for the theta direction
  for (i=0; i<Data[35]; i++)
  {

    // loop for the phi direction
    for (j=0; j<PhiNum; j++)
    {

      // print cos(theta) for a model space symmetric to the xy-plane
      if (Data[37]==1)
      {
         res_stream << "cosT " << 1 - (i+0.5)/Data[35];
      }

      // print cos(theta) for a model space not symmetric to the
      // xy-plane
      else
      {
         res_stream << "cosT " << 1 - 2*(i+0.5)/Data[35];
      }

      // if there are more than 1 azimuthal viewing angles print 
      // the value of phi
      if (PhiNum>1)
      {
        res_stream << "/P" << (j+0.5)*360.0/Data[34] << " ";
      }
	else
      {
        res_stream << " ";
      }

      out_stream << "\n";
      res_stream << "\n";

      // spectral loop
      if(Data[39]==1 && Data[40]==1)
      {
		x=0;
		y=0;

  		dk = Data[4]/Data[36];		
		k=Data[2]+dk/2;			

     		// loop for the spectroscopic channels
      		for (l=0; l<Data[36]; l++)
      		{
        		I[l][x][y] = Col[i].Row[j].I[l][x][y];
        		Q[l][x][y] = Col[i].Row[j].Q[l][x][y];
        		U[l][x][y] = Col[i].Row[j].U[l][x][y];
        		V[l][x][y] = Col[i].Row[j].V[l][x][y];
		

			// compute the polarization p
        		if (I[l][x][y]!=0)
        		{
          		  p = sqrt(Q[l][x][y]*Q[l][x][y]+U[l][x][y]*U[l][x][y]+V[l][x][y]*V[l][x][y])/I[l][x][y];
        		}
        		else
        		{
        		  p = 0;
        		}

        		// compute the polarization position angle
        		if ((U[l][x][y]==0)&&(Q[l][x][y]==0))
        		{
        		  theta = 0;
        		}
        		else
        		{
        		  theta = 90/pi*atan2(U[l][x][y],Q[l][x][y]);
        		}

			// compute a normalization constant for the fluxes
        		N = (Data[0]*(Data[2]+(l+0.5)*Data[4]/Data[36]))/10000;

			out_stream << k << " ";	
			res_stream << k << " ";	
			k=k+dk;			

        		switch(n)
        		{

        		  // save combined I-parameter and normalized total flux
        		  case 0:
         		    out_stream << I[l][x][y] << " ";
         		    res_stream << I[l][x][y]/N << " ";

			  // save combined Q-parameter and polarization
        		  case 1:
        		    out_stream << Q[l][x][y] << " ";
         		    res_stream << p << " ";
         		    break;

	  		  // save combined U-parameter and position angle
         		  case 2:
          		    out_stream << U[l][x][y] << " ";
         		    if (theta < -45) theta = theta + 180;
        		    res_stream << theta << " ";
         		    break;

			  // save combined V parameter and normalized polarized flux
        		  default:
        		    out_stream << V[l][x][y] << " ";
        		    res_stream << p*I[l][x][y]/N << " ";
        		    break;
        		}

      			out_stream << "\n";
      			res_stream << "\n";
		
		}
      }
      // imagining loop
      else
      {
       for (x=0; x<Data[39]; x++)
       {
	 for (y=0; y<Data[40]; y++)
	 {      
     		// loop for the spectroscopic channels
      		for (l=0; l<Data[36]; l++)
      		{
        		I[l][x][y] = Col[i].Row[j].I[l][x][y];
        		Q[l][x][y] = Col[i].Row[j].Q[l][x][y];
        		U[l][x][y] = Col[i].Row[j].U[l][x][y];
        		V[l][x][y] = Col[i].Row[j].V[l][x][y];
		

			// compute the polarization p
        		if (I[l][x][y]!=0)
        		{
          		  p = sqrt(Q[l][x][y]*Q[l][x][y]+U[l][x][y]*U[l][x][y]+V[l][x][y]*V[l][x][y])/I[l][x][y];
        		}
        		else
        		{
        		  p = 0;
        		}

        		// compute the polarization position angle
        		if ((U[l][x][y]==0)&&(Q[l][x][y]==0))
        		{
        		  theta = 0;
        		}
        		else
        		{
        		  theta = 90/pi*atan2(U[l][x][y],Q[l][x][y]);
        		}

			// compute a normalization constant for the fluxes
        		N = (Data[0]*(Data[2]+(l+0.5)*Data[4]/Data[36]))/10000;

			// print X and Y value
      			out_stream << -((Data[41]/2)-0.5*(Data[41]/Data[39]))+x*(Data[41]/Data[39]) << " ";
      			res_stream << -((Data[41]/2)-0.5*(Data[41]/Data[39]))+x*(Data[41]/Data[39]) << " ";
      			out_stream << -((Data[41]/2)-0.5*(Data[41]/Data[40]))+y*(Data[41]/Data[40]) << " ";
      			res_stream << -((Data[41]/2)-0.5*(Data[41]/Data[40]))+y*(Data[41]/Data[40]) << " ";

        		switch(n)
        		{

        		  // save combined I-parameter and normalized total flux
        		  case 0:
         		    out_stream << I[l][x][y] << " ";
         		    res_stream << I[l][x][y]/N << " ";
                            break;

			  // save combined Q-parameter and polarization
        		  case 1:
        		    out_stream << Q[l][x][y] << " ";
         		    res_stream << p << " ";
         		    break;

	  		  // save combined U-parameter and position angle
         		  case 2:
          		    out_stream << U[l][x][y] << " ";
         		    if (theta < -45) theta = theta + 180;
        		    res_stream << theta << " ";
         		    break;

			  // save combined V parameter and normalized polarized flux
        		  default:
        		    out_stream << V[l][x][y] << " ";
        		    res_stream << p*I[l][x][y]/N << " ";
        		    break;
        		}

      			out_stream << "\n";
      			res_stream << "\n";
		
		}

        }
       }
      }
      out_stream << "\n";
      res_stream << "\n";
  }
 }

  out_stream.close();
  res_stream.close();
}
